/* ruler97.c */

/* rulerMarks integer benchmark program
 * (K) All Rites Reversed -- Copy What You Like
 *
 * Author:
 *      Peter Hartley <peter@ant.co.uk>
 *
 * See also:
 *      http://utter.chaos.org.uk/~pdh/software/marks.htm
 */

/* 64-bit Goulomb ruler code */

#include <stdio.h>
#include <time.h>
#include <stdlib.h>

#ifdef __unix__
#include <sys/time.h>

struct timeval begin, end;
#endif

/* ------------------------------------------------------------------------ */
/* Valid MARKS are 8,9,10,11                                                */

#ifndef MARKS
#define MARKS 11
#endif

/* maximum gap tried */

#define MAX 20

/* name of results file */

#define OUTPUTFILE "rulerout"

/* ------------------------------------------------------------------------ */

FILE *fff = NULL;

void bin( int i )
{
    while(i) {
        putchar('0' + (i&1));
        if ( fff )
            putc( '0' + (i&1), fff );
        i = i >> 1;
    };
}

int num[64],
    best[64][10];

void show( int i )
{
    int j,k;
    time_t tnow;

    if ( fff )
        fprintf( fff, "f(%2d) = %d (%3d of) 1", i, MARKS, num[i] );

    for ( k=0; k < MARKS-1; k++ )
    {
        for (j=1; j<best[i][k]; j++) { putc( 48, fff ); }
        putc( 49, fff );
    }
    tnow = time(0);
    fprintf( fff, " %s", ctime( &tnow ) );
    fflush( fff );
}

/* rulermark stuff */
void rulermark(void)
{
#ifdef __unix__
    gettimeofday(&end, NULL);
    double dt = (end.tv_sec - begin.tv_sec) + (end.tv_usec - begin.tv_usec)*0.000001;
#else
    int t = clock();
    double dt = ((double)t)/CLOCKS_PER_SEC;
#endif

    printf( "Time for rulerMarks benchmark = %.3fs\n", dt );
    printf( "rulerMarks97 rating = %.1f\n", 207.11/dt );
    exit(0);
}

#ifndef min
#define min(x,y) ((x>y)?y:x)
#endif

#define CALC_NEXT(lo,hi)            \
        if ( lo == 0xFFFFFFFF )     \
            { v = hi; jj=32; }      \
        else                        \
            { v = lo; jj=0; }       \
        while (v & 1)               \
            { v = v >> 1; jj++; }

int main( int argc, char *argv[] )
{
    int a,b,c,d=0,e=0,f=0,g=0,h=0,i=0,j=0;
    unsigned int max = 0;
    unsigned int jj, v, vhi;
    unsigned int f1,f2,f3,f4,f5,f6,f7,f8,f9,f10;
    unsigned int f1hi,f2hi,f3hi,f4hi,f5hi,f6hi,f7hi,f8hi,f9hi,f10hi;
    unsigned int v1,v2,v3,v4,v5,v6,v7,v8,v9,v10;
    unsigned int v1hi,v2hi,v3hi,v4hi,v5hi,v6hi,v7hi,v8hi,v9hi,v10hi;
    int e3,e4,e5,e6,e7,e8,e9;
    time_t tnow;

#ifdef __unix__
    gettimeofday(&begin, NULL);
#endif

    fff = fopen( OUTPUTFILE, "w" );

    if ( fff == 0 )
    {
        puts( "Cannot open output file" );
        return 1;
    }

    for (jj=0; j<64; j++)
        num[jj] = 0;

    f1 = 1; f1hi = 0;
    for (a=1; a<MAX; a++)
    {
        v1 = f1; v1hi = f1hi;
        f2 = (f1<<1) + 1;
        f2hi = (f1hi<<1) + (f1>>31);
        for (b=1; b<MAX; b++)
        {
            char * res;
            tnow = time(0);
            res = ctime( &tnow );
            printf("  %d/%d done, time is %s", (a-1)*(MAX-1)+(b-1),
                                               (MAX-1)*(MAX-1), res );
            fprintf( fff, "  %d/%d done, time is %s", (a-1)*(MAX-1)+(b-1),
                                               (MAX-1)*(MAX-1), res );
            fflush( fff );
            v2 = v1 | f2;
            v2hi = v1hi | f2hi;

            f3 = (f2<<1) + 1;
            f3hi = (f2hi<<1) + (f2>>31);
            for (c=1; c<MAX; c++)
            {
                v3 = v2 | f3;
                v3hi = v2hi | f3hi;

                e3 = 64-a-b-c-MARKS+2;

                f4   = (f3  <<1) + 1;
                f4hi = (f3hi<<1) + (f3>>31);
                for (d=1; d<min(e3,MAX); d++)
                {
                    v4   = v3   | f4;
                    v4hi = v3hi | f4hi;
#if MARKS >= 6
                    e4=e3-d;
                    f5   = (f4  <<1) + 1;
                    f5hi = (f4hi<<1) + (f4>>31);
                    for (e=1; e<min(e4,MAX); e++)
                    {
                        v5   = v4   | f5;
                        v5hi = v4hi | f5hi;
#endif
#if MARKS >= 7
                        e5=e4-e;
                        f6   = (f5  <<1) + 1;
                        f6hi = (f5hi<<1) + (f5>>31);
                        for (f=1; f<min(e5,MAX); f++)
                        {
                            v6   = v5   | f6;
                            v6hi = v5hi | f6hi;
#endif
#if MARKS >= 8
                            e6=e5-f;
                            f7   = (f6  <<1) + 1;
                            f7hi = (f6hi<<1) + (f6>>31);
                            for (g=1; g<min(e6,MAX); g++)
                            {
                                v7   = v6   | f7;
                                v7hi = v6hi | f7hi;
#endif
#if MARKS >= 9
                                e7=e6-g;
                                f8   = (f7  <<1) + 1;
                                f8hi = (f7hi<<1) + (f7>>31);
                                for (h=1; h<min(e7,MAX); h++)
                                {
                                    v8   = v7   | f8;
                                    v8hi = v7hi | f8hi;
#endif
#if MARKS >= 10
                                    e8 = e7-h;
                                    f9   = (f8  <<1) + 1;
                                    f9hi = (f8hi<<1) + (f8>>31);
                                    for (i=1; i<min(e8,MAX); i++)
                                    {
                                        v9   = v8   | f9;
                                        v9hi = v8hi | f9hi;
#endif
#if MARKS >= 11
                                        e9 = e8-i;
# if MARKS == 11 && defined(FLIP)
                                        CALC_NEXT( v9, v9hi );
                                        e9 = min(e9,jj+1);
# endif
                                        f10   = (f9  <<1) + 1;
                                        f10hi = (f9hi<<1) + (f9>>31);
                                        for (j=1; j<min(e9,MAX); j++)
                                        {
                                            v10   = v9   | f10;
                                            v10hi = v9hi | f10hi;
#endif

                                            /*
                                            printf("%d,%d,%d,%d => ",a,b,c,d);
                                            bin(f4); putchar(' ');
                                            bin(v4); putchar('\n');
                                            */
#if MARKS == 11
                                       v = v10;
                                       vhi = v10hi;
#else
 #if MARKS == 10
                                       v = v9;
                                       vhi = v9hi;
 #else
  #if MARKS == 9
                                       v = v8;
                                       vhi = v8hi;
  #else
   #if MARKS == 8
                                       v = v7;
                                       vhi = v7hi;
   #else
    #if MARKS == 7
                                       v = v6;
                                       vhi = v6hi;
    #else
     #if MARKS == 6
                                       v = v5;
                                       vhi = v5hi;
     #else
                                       v = v4;
                                       vhi = v4hi;
     #endif
    #endif
   #endif
  #endif
 #endif
#endif

                                            if (v == 0xFFFFFFFF)
                                            {
                                                jj = 32;
                                                v = vhi;
                                            }
                                            else
                                                jj = 0;
                                            while (v & 1)
                                            {
                                                v = v >> 1;
                                                jj++;
                                            }
                                            num[jj] = num[jj] + 1;
                                            if ( num[jj] == 1 || jj >= max )
                                            {
                                                best[jj][0]=a;
                                                best[jj][1]=b;
                                                best[jj][2]=c;
                                                best[jj][3]=d;
                                                best[jj][4]=e;
                                                best[jj][5]=f;
                                                best[jj][6]=g;
                                                best[jj][7]=h;
                                                best[jj][8]=i;
                                                best[jj][9]=j;

                                                show(jj);

                                                /* rulermark stuff */
                                                if ( jj == 3
                                                     && num[jj] == 1)
                                                    rulermark();
                                                /* end of rulermark stuff */

                                                if ( jj >= max )
                                                    max = jj;
                                            }
#if MARKS >= 11
                                            f10hi += f10hi + (f10>>31);
                                            f10   += f10;
                                        }
#endif
#if MARKS >= 10
                                        f9hi += f9hi + (f9>>31);
                                        f9 += f9;
                                    }
#endif
#if MARKS >= 9
                                    f8hi += f8hi + (f8>>31);
                                    f8 += f8;
                                }
#endif
#if MARKS >= 8
                                f7hi += f7hi + (f7>>31);
                                f7 += f7;
                            }
#endif
#if MARKS >= 7
                            f6hi += f6hi + (f6>>31);
                            f6 += f6;
                        }
#endif
#if MARKS >= 6
                        f5hi += f5hi + (f5>>31);
                        f5 += f5;
                    }
#endif
                    f4hi += f4hi + (f4>>31);
                    f4 += f4;
                }
                f3hi += f3hi + (f3>>31);
                f3 += f3;
            }
            f2hi += f2hi + (f2>>31);
            f2 += f2;
        }
        f1hi += f1hi + (f1>>31);
        f1 += f1;
    }
    for (jj=0; jj<64; jj++)
    {
        if (num[jj] > 0) show(jj);
    }
    fclose( fff );
    return 0;
}
